@@ -8,6 +8,7 @@ Description: Functions to create new Pandas DataFrame features from existing dat
 import pandas as pd
 import numpy as np
 import itertools
+from datetime import datetime, timedelta
 
 
 def cols_to_log(df, columns):
@@ -303,54 +304,6 @@ def get_previous_value(df, group, column, name):
     return df
 
 
-def get_days_since_date(df, before_datetime, after_datetime, name):
-    """Return a new column containing the difference between two dates in days.
-
-    Args:
-        df: Pandas DataFrame.
-        before_datetime: Earliest datetime (will convert value)
-        after_datetime: Latest datetime (will convert value)
-        name: Name for new column.
-
-    Returns:
-        Original DataFrame with new column containing previous value of named column.
-
-    """
-    df = df.copy()
-    df[before_datetime] = pd.to_datetime(df[before_datetime])
-    df[after_datetime] = pd.to_datetime(df[after_datetime])
-
-    df[name] = df[after_datetime] - df[before_datetime]
-    df[name] = df[name] / np.timedelta64(1, 'D')
-
-    return df
-
-
-def get_dates(df, date_column):
-    """Converts a given date to various formats and returns an updated DataFrame.
-
-    Args:
-        df: Pandas DataFrame.
-        date_column:
-
-    Returns:
-        Original DataFrame with additional date columns.
-    """
-
-    df['day'] = df[date_column].dt.strftime("%d")  # Day of month with leading zero
-    df['month'] = df[date_column].dt.strftime("%m")  # Month of year with leading zero
-    df['year'] = df[date_column].dt.strftime("%Y")  # Full numeric four digit year
-    df['year_month'] = df[date_column].dt.strftime("%Y%m")  # Full numeric four digit year plus month
-    df['week_number'] = df[date_column].dt.strftime("%U")  # Week number with leading zero
-    df['day_number'] = df[date_column].dt.strftime("%j")  # Day number with leading zero
-    df['day_name'] = df[date_column].dt.strftime("%A")  # Day name, i.e. Sunday
-    df['month_name'] = df[date_column].dt.strftime("%B")  # Month name, i.e. January
-    df['mysql_date'] = df[date_column].dt.strftime("%Y-%d-%m")  # MySQL date, i.e. 2020-30-01
-    df['quarter'] = df[date_column].dt.quarter  # Quarter with leading zero, i.e. 01
-
-    return df
-
-
 def get_grouped_stats(df, group, columns):
     """Group by a column and return summary statistics for a list of columns and add prefixed data to new columns.
 
@@ -644,3 +597,192 @@ def get_conversion_rate(df, total, conversions):
     value = (df[conversions] / df[total])
     return value
 
+
+def count_subzero(df):
+    """Count the number of subzero values in a Pandas DataFrame row.
+    If you fill NaN values with -1 during feature engineering, instead
+    of zero, you can use this to create a new feature, which may yield
+    more valuable information than the mere presence of zeros.
+
+    Args:
+        :param df: Pandas DataFrame.
+
+    Returns:
+        Column value for each row in DataFrame
+
+    Example:
+
+        df = count_subzero(df)
+    """
+
+    return np.sum(df < 0, axis=1)
+
+
+def count_zero(df):
+    """Count the number of zero values in a Pandas DataFrame row to create a new feature.
+
+    Args:
+        :param df: Pandas DataFrame.
+
+    Returns:
+        Column value for each row in DataFrame
+
+    Example:
+
+        df = count_zero(df)
+    """
+
+    return np.sum(df == 0, axis=1)
+
+
+def count_missing(df):
+    """Count the number of missing or NaN values in a Pandas DataFrame row to create a new feature.
+
+    Args:
+        :param df: Pandas DataFrame.
+
+    Returns:
+        Column value for each row in DataFrame
+
+    Example:
+
+        df = count_missing(df)
+    """
+
+    return df.isnull().sum(axis=1)
+
+
+def get_grouped_metric(df, group_column, metric_column, operation):
+    """Group by a specified column, then perform a mathematical operation on a given
+    metric column and return the value so it can be assigned to the DataFrame.
+    For example, group by ID and sum the value of orders placed by the ID.
+
+    :param df: Pandas DataFrame
+    :param group_column: Column name to use for groupby, i.e. id
+    :param metric_column: Column name for metric column, i.e. visits
+    :param operation: Operation to perform  (sum, count, nunique, min, max, median, mean)
+    """
+
+    return df.groupby(group_column)[metric_column].transform(operation)
+
+
+def get_grouped_metric_where(df, group_column, operation, metric_column, where_column, where_operator, where_value):
+    """Groups a Pandas DataFrame by a specified column, then performs a mathematical operation on a column
+    based on a where value and clause. For example, group by ID and count sessions where productAddToCarts was
+    >= 1.
+
+    :param df: Pandas DataFrame
+    :param group_column: Column to group by, i.e. ID
+    :param operation: Mathematical operation to perform (sum, count, nunique, min, max, median, mean)
+    :param metric_column: Column to use in metric operation, i.e. sessions
+    :param where_column: Column name to use in where clause, i.e. productAddToCarts
+    :param where_operator: Where operator to use in clause (i.e. eq, ne, le, ge, lt, gt)
+    :param where_value: Where clause value, i.e. 1
+
+    Examples:
+    df['engagements'] = get_grouped_metric_where(df, 'id', 'count', 'sessions', 'productAddsToCart', 'ge', 1)
+
+    """
+
+    if where_operator == 'eq':
+        return df[df[where_column] == where_value].groupby(group_column)[metric_column].transform(operation)
+
+    if where_operator == 'ne':
+        return df[df[where_column] != where_value].groupby(group_column)[metric_column].transform(operation)
+
+    if where_operator == 'le':
+        return df[df[where_column] <= where_value].groupby(group_column)[metric_column].transform(operation)
+
+    if where_operator == 'ge':
+        return df[df[where_column] >= where_value].groupby(group_column)[metric_column].transform(operation)
+
+    if where_operator == 'lt':
+        return df[df[where_column] < where_value].groupby(group_column)[metric_column].transform(operation)
+
+    if where_operator == 'gt':
+        return df[df[where_column] > where_value].groupby(group_column)[metric_column].transform(operation)
+
+    else:
+        return 0
+
+
+def date_subtract(date, days_to_subtract, date_format):
+    """Subtracts a given number of days from a date and returns the date in a defined format.
+
+    :param date: Python datetime value from which to subtract
+    :param days_to_subtract: Number of days to subtract from the date
+    :param date_format: Date format to return (i.e. '%Y-%m-%d' for 2020-06-30)
+    :returns: Date in specific format less X days
+
+    Examples:
+        df['date_minus_7_days'] = date_subtract(df['date'], 7, '%Y-%m-%d')
+        df['date_30_days_ago'] = date_subtract(datetime.today(), 30, '%Y-%m-%d')
+    """
+
+    return (date - timedelta(days=days_to_subtract)).strftime(date_format)
+
+
+def date_add(date, days_to_add, date_format):
+    """Adds a given number of days to a date and returns the date in a defined format.
+
+    :param date: Python datetime value on which to add
+    :param days_to_add: Number of days to add to date
+    :param date_format: Date format to return (i.e. '%Y-%m-%d' for 2020-06-30)
+    :returns: Date in specific format plus X days
+
+    Examples:
+        df['today_plus_7_days'] = date_add(df['date'], 7, '%Y-%m-%d')
+        df['date_plus_30_days'] = date_add(datetime.today(), 30, '%Y-%m-%d')
+    """
+
+    return (date + timedelta(days=days_to_add)).strftime(date_format)
+
+
+def get_days_since_date(df, before_datetime, after_datetime, name):
+    """Return a new column containing the difference between two dates in days.
+
+    Args:
+        df: Pandas DataFrame.
+        before_datetime: Earliest datetime (will convert value)
+        after_datetime: Latest datetime (will convert value)
+        name: Name for new column.
+
+    Returns:
+        Original DataFrame with new column containing previous value of named column.
+
+    """
+    df = df.copy()
+    df[before_datetime] = pd.to_datetime(df[before_datetime])
+    df[after_datetime] = pd.to_datetime(df[after_datetime])
+
+    df[name] = df[after_datetime] - df[before_datetime]
+    df[name] = df[name] / np.timedelta64(1, 'D')
+
+    return df
+
+
+def get_dates(df, date_column):
+    """Converts a given date to various formats and returns an updated DataFrame.
+
+    Args:
+        df: Pandas DataFrame.
+        date_column:
+
+    Returns:
+        Original DataFrame with additional date columns.
+    """
+
+    df['day'] = df[date_column].dt.strftime("%d").astype(int)  # Day of month with leading zero
+    df['month'] = df[date_column].dt.strftime("%m").astype(int)  # Month of year with leading zero
+    df['year'] = df[date_column].dt.strftime("%Y").astype(int)  # Full numeric four digit year
+    df['year_month'] = df[date_column].dt.strftime("%Y%m").astype(int)  # Full numeric four digit year plus month
+    df['week_number'] = df[date_column].dt.strftime("%U").astype(int)  # Week number with leading zero
+    df['day_number'] = df[date_column].dt.strftime("%j").astype(int)  # Day number with leading zero
+    df['day_name'] = df[date_column].dt.strftime("%A")  # Day name, i.e. Sunday
+    df['month_name'] = df[date_column].dt.strftime("%B")  # Month name, i.e. January
+    df['mysql_date'] = df[date_column].dt.strftime("%Y-%d-%m")  # MySQL date, i.e. 2020-30-01
+    df['quarter'] = df[date_column].dt.quarter.astype(int)  # Quarter with leading zero, i.e. 01
+    df['week_day_number'] = df[date_column].dt.strftime("%w").astype(int)  # Weekday number, i.e. 0 = Sunday, 1 = Monday
+
+    return df
+
